@@ -10,8 +10,10 @@ import (
 	"github.com/google/uuid"
 	machinev1 "github.com/openshift/machine-api-operator/pkg/apis/machine/v1beta1"
 	vspherev1 "github.com/openshift/machine-api-operator/pkg/apis/vsphereprovider/v1beta1"
+	machineapierros "github.com/openshift/machine-api-operator/pkg/controller/machine"
 	machinecontroller "github.com/openshift/machine-api-operator/pkg/controller/machine"
 	"github.com/openshift/machine-api-operator/pkg/controller/vsphere/session"
+	"github.com/openshift/machine-api-operator/pkg/metrics"
 	"github.com/vmware/govmomi/find"
 	"github.com/vmware/govmomi/object"
 	"github.com/vmware/govmomi/property"
@@ -21,7 +23,7 @@ import (
 	"github.com/vmware/govmomi/vim25/mo"
 	"github.com/vmware/govmomi/vim25/types"
 	corev1 "k8s.io/api/core/v1"
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 )
 
 const (
@@ -60,43 +62,53 @@ func (r *Reconciler) create() error {
 		return fmt.Errorf("%v: failed validating machine provider spec: %v", r.machine.GetName(), err)
 	}
 
-	if r.providerStatus.TaskRef != "" {
-		moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+	// We only clone the VM template if we have no taskRef.
+	if r.providerStatus.TaskRef == "" {
+		if !r.machineScope.session.IsVC() {
+			return fmt.Errorf("%v: not connected to a vCenter", r.machine.GetName())
+		}
+		klog.Infof("%v: cloning", r.machine.GetName())
+		task, err := clone(r.machineScope)
 		if err != nil {
-			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
-				return err
-			}
-		}
-		if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
-			if !taskIsFinished {
-				return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+			conditionFailed := conditionFailed()
+			conditionFailed.Message = err.Error()
+			statusError := setProviderStatus(task, conditionFailed, r.machineScope, nil)
+			if statusError != nil {
+				return fmt.Errorf("Failed to set provider status: %w", err)
 			}
 			return err
 		}
+		return setProviderStatus(task, conditionSuccess(), r.machineScope, nil)
 	}
 
-	if _, err := findVM(r.machineScope); err != nil {
-		if !isNotFound(err) {
-			return err
-		}
-		if r.machineScope.session.IsVC() {
-			klog.Infof("%v: cloning", r.machine.GetName())
-			task, err := clone(r.machineScope)
-			if err != nil {
-				conditionFailed := conditionFailed()
-				conditionFailed.Message = err.Error()
-				statusError := setProviderStatus(task, conditionFailed, r.machineScope, nil)
-				if statusError != nil {
-					return fmt.Errorf("Failed to set provider status: %w", err)
-				}
-				return err
-			}
-
-			return setProviderStatus(task, conditionSuccess(), r.machineScope, nil)
-		}
-		return fmt.Errorf("%v: not connected to a vCenter", r.machine.GetName())
+	moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+	if err != nil {
+		return err
 	}
 
+	if moTask == nil {
+		// Possible eventual consistency problem from vsphere
+		// TODO: change error message here to indicate this might be expected.
+		return fmt.Errorf("Unexpected moTask nil")
+	}
+
+	// TODO: this should be handled on the machine-controller side, we should
+	// pass an specific error type for this case.
+	// TODO: this should probably be terminal and we should set failed.
+	if moTask.Info.State == types.TaskInfoStateError {
+		metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+			Name:      r.machine.Name,
+			Namespace: r.machine.Namespace,
+			Reason:    fmt.Sprintf("Create machine task finished with error: %+v", moTask.Info.Error),
+		})
+	}
+	if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
+		if !taskIsFinished {
+			return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+		}
+		return err
+	}
+	// If taskIsFinished then next reconcile should result in update.
 	return nil
 }
 
@@ -107,17 +119,26 @@ func (r *Reconciler) update() error {
 	}
 
 	if r.providerStatus.TaskRef != "" {
-		motask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
+		moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
 		if err != nil {
 			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
 				return err
 			}
 		}
-		if taskIsFinished, err := taskIsFinished(motask); err != nil || !taskIsFinished {
-			if !taskIsFinished {
-				return fmt.Errorf("task %v has not finished", motask.Reference().Value)
+		if moTask != nil {
+			if moTask.Info.State == types.TaskInfoStateError {
+				metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+					Name:      r.machine.Name,
+					Namespace: r.machine.Namespace,
+					Reason:    fmt.Sprintf("Update machine task finished with error: %+v", moTask.Info.Error),
+				})
+			}
+			if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
+				if !taskIsFinished {
+					return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+				}
+				return err
 			}
-			return err
 		}
 	}
 
@@ -161,23 +182,40 @@ func (r *Reconciler) exists() (bool, error) {
 
 func (r *Reconciler) delete() error {
 	if r.providerStatus.TaskRef != "" {
+		// TODO: We need to use a separate status field for the create and the
+		// delete taskref.
 		moTask, err := r.session.GetTask(r.Context, r.providerStatus.TaskRef)
 		if err != nil {
 			if !isRetrieveMONotFound(r.providerStatus.TaskRef, err) {
 				return err
 			}
 		}
-		if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
-			if !taskIsFinished {
-				return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+		if moTask != nil {
+			if moTask.Info.State == types.TaskInfoStateError {
+				// TODO: this should be DeleteInstance event.
+				metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+					Name:      r.machine.Name,
+					Namespace: r.machine.Namespace,
+					Reason:    fmt.Sprintf("Delete machine task finished with error: %+v", moTask.Info.Error),
+				})
+			}
+			if taskIsFinished, err := taskIsFinished(moTask); err != nil || !taskIsFinished {
+				if !taskIsFinished {
+					return fmt.Errorf("task %v has not finished", moTask.Reference().Value)
+				}
+				return err
 			}
-			return err
 		}
 	}
 
 	vmRef, err := findVM(r.machineScope)
 	if err != nil {
 		if !isNotFound(err) {
+			metrics.RegisterFailedInstanceDelete(&metrics.MachineLabels{
+				Name:      r.machine.Name,
+				Namespace: r.machine.Namespace,
+				Reason:    err.Error(),
+			})
 			return err
 		}
 		klog.Infof("%v: vm does not exist", r.machine.GetName())
@@ -194,8 +232,15 @@ func (r *Reconciler) delete() error {
 		return err
 	}
 
+	// Will this error if VM not all the way powered off yet?  Don't want to
+	// emit an event for a transient condition.
 	task, err := vm.Obj.Destroy(r.Context)
 	if err != nil {
+		metrics.RegisterFailedInstanceDelete(&metrics.MachineLabels{
+			Name:      r.machine.Name,
+			Namespace: r.machine.Namespace,
+			Reason:    err.Error(),
+		})
 		return fmt.Errorf("%v: failed to destroy vm: %v", r.machine.GetName(), err)
 	}
 
@@ -203,6 +248,7 @@ func (r *Reconciler) delete() error {
 		return fmt.Errorf("Failed to set provider status: %w", err)
 	}
 
+	// TODO: consider returning an error to specify retry time here
 	return fmt.Errorf("destroying vm in progress, reconciling")
 }
 
@@ -258,6 +304,11 @@ func (r *Reconciler) reconcileRegionAndZoneLabels(vm *virtualMachine) error {
 	})
 
 	if err != nil {
+		metrics.RegisterFailedInstanceUpdate(&metrics.MachineLabels{
+			Name:      r.machine.Name,
+			Namespace: r.machine.Namespace,
+			Reason:    err.Error(),
+		})
 		return err
 	}
 
@@ -410,7 +461,9 @@ func clone(s *machineScope) (string, error) {
 
 	vmTemplate, err := s.GetSession().FindVM(*s, "", s.providerSpec.Template)
 	if err != nil {
-		return "", err
+		const notFoundMsg = "template not found, specify valid value"
+		defaultError := fmt.Errorf("unable to get template %q: %w", s.providerSpec.Template, err)
+		return "", handleVSphereError("", notFoundMsg, defaultError, err)
 	}
 
 	var snapshotRef *types.ManagedObjectReference
@@ -456,24 +509,24 @@ func clone(s *machineScope) (string, error) {
 
 	folder, err := s.GetSession().Finder.FolderOrDefault(s, folderPath)
 	if err != nil {
-		multipleFoundMsg := "multiple folders found, specify one in config"
-		notFoundMsg := "folder not found, specify valid value"
+		const multipleFoundMsg = "multiple folders found, specify one in config"
+		const notFoundMsg = "folder not found, specify valid value"
 		defaultError := fmt.Errorf("unable to get folder for %q: %w", folderPath, err)
 		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
 
 	datastore, err := s.GetSession().Finder.DatastoreOrDefault(s, datastorePath)
 	if err != nil {
-		multipleFoundMsg := "multiple datastores found, specify one in config"
-		notFoundMsg := "datastore not found, specify valid value"
+		const multipleFoundMsg = "multiple datastores found, specify one in config"
+		const notFoundMsg = "datastore not found, specify valid value"
 		defaultError := fmt.Errorf("unable to get datastore for %q: %w", datastorePath, err)
 		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
 
 	resourcepool, err := s.GetSession().Finder.ResourcePoolOrDefault(s, resourcepoolPath)
 	if err != nil {
-		multipleFoundMsg := "multiple resource pools found, specify one in config"
-		notFoundMsg := "resource pool not found, specify valid value"
+		const multipleFoundMsg = "multiple resource pools found, specify one in config"
+		const notFoundMsg = "resource pool not found, specify valid value"
 		defaultError := fmt.Errorf("unable to get resource pool for %q: %w", resourcepool, err)
 		return "", handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 	}
@@ -550,11 +603,17 @@ func clone(s *machineScope) (string, error) {
 
 	task, err := vmTemplate.Clone(s, folder, s.machine.GetName(), spec)
 	if err != nil {
-		return "", fmt.Errorf("error triggering clone op for machine %v: %w", s, err)
+		err = fmt.Errorf("error triggering clone op for machine %v: %w", s, err)
+		metrics.RegisterFailedInstanceCreate(&metrics.MachineLabels{
+			Name:      s.machine.Name,
+			Namespace: s.machine.Namespace,
+			Reason:    err.Error(),
+		})
+		return "", err
 	}
-
-	klog.V(3).Infof("%v: running task: %+v", s.machine.GetName(), s.providerStatus.TaskRef)
-	return task.Reference().Value, nil
+	taskVal := task.Reference().Value
+	klog.V(3).Infof("%v: running task: %+v", s.machine.GetName(), taskVal)
+	return taskVal, nil
 }
 
 func getDiskSpec(s *machineScope, devices object.VirtualDeviceList) (types.BaseVirtualDeviceConfigSpec, error) {
@@ -564,7 +623,13 @@ func getDiskSpec(s *machineScope, devices object.VirtualDeviceList) (types.BaseV
 	}
 
 	disk := disks[0].(*types.VirtualDisk)
-	disk.CapacityInKB = int64(s.providerSpec.DiskGiB) * 1024 * 1024
+	cloneCapacityKB := int64(s.providerSpec.DiskGiB) * 1024 * 1024
+	if disk.CapacityInKB > cloneCapacityKB {
+		return nil, machineapierros.InvalidMachineConfiguration(
+			"can't resize template disk down, initial capacity is larger: %dKiB > %dKiB",
+			disk.CapacityInKB, cloneCapacityKB)
+	}
+	disk.CapacityInKB = cloneCapacityKB
 
 	return &types.VirtualDeviceConfigSpec{
 		Operation: types.VirtualDeviceConfigSpecOperationEdit,
@@ -589,7 +654,10 @@ func getNetworkDevices(s *machineScope, devices object.VirtualDeviceList) ([]typ
 
 		ref, err := s.GetSession().Finder.Network(s.Context, netSpec.NetworkName)
 		if err != nil {
-			return nil, fmt.Errorf("unable to find network %q: %w", netSpec.NetworkName, err)
+			const multipleFoundMsg = "multiple networks found, specify one in config"
+			const notFoundMsg = "network not found, specify valid value"
+			defaultError := fmt.Errorf("unable to get network for %q: %w", netSpec.NetworkName, err)
+			return nil, handleVSphereError(multipleFoundMsg, notFoundMsg, defaultError, err)
 		}
 
 		backing, err := ref.EthernetCardBackingInfo(s.Context)
@@ -643,7 +711,7 @@ func taskIsFinished(task *mo.Task) (bool, error) {
 	case types.TaskInfoStateSuccess:
 		return true, nil
 	case types.TaskInfoStateError:
-		return true, nil
+		return true, errors.New(task.Info.Error.LocalizedMessage)
 	default:
 		return false, fmt.Errorf("task: %v, unknown state %v", task.Reference().Value, task.Info.State)
 	}
@@ -813,6 +881,11 @@ func (vm *virtualMachine) reconcileTags(ctx context.Context, session *session.Se
 			klog.Infof("%v: Attaching %s tag to vm", machine.GetName(), clusterID)
 			// the tag should already be created by installer
 			if err := m.AttachTag(ctx, clusterID, vm.Ref); err != nil {
+				metrics.RegisterFailedInstanceUpdate(&metrics.MachineLabels{
+					Name:      machine.Name,
+					Namespace: machine.Namespace,
+					Reason:    err.Error(),
+				})
 				return err
 			}
 		}
@@ -825,8 +898,18 @@ func (vm *virtualMachine) reconcileTags(ctx context.Context, session *session.Se
 	return nil
 }
 
-// checkAttachedTag returns true if tag is already attached to a vm
+// checkAttachedTag returns true if tag is already attached to a vm or tag doesn't exist
 func (vm *virtualMachine) checkAttachedTag(ctx context.Context, tagName string, m *tags.Manager) (bool, error) {
+	// cluster ID tag doesn't exists in UPI, we should skip tag attachment if it's not found
+	foundTag, err := vm.foundTag(ctx, tagName, m)
+	if err != nil {
+		return false, err
+	}
+
+	if !foundTag {
+		return true, nil
+	}
+
 	tags, err := m.GetAttachedTags(ctx, vm.Ref)
 	if err != nil {
 		return false, err
@@ -841,6 +924,26 @@ func (vm *virtualMachine) checkAttachedTag(ctx context.Context, tagName string,
 	return false, nil
 }
 
+func (vm *virtualMachine) foundTag(ctx context.Context, tagName string, m *tags.Manager) (bool, error) {
+	tags, err := m.ListTags(ctx)
+	if err != nil {
+		return false, err
+	}
+
+	for _, id := range tags {
+		tag, err := m.GetTag(ctx, id)
+		if err != nil {
+			return false, err
+		}
+
+		if tag.Name == tagName {
+			return true, nil
+		}
+	}
+
+	return false, nil
+}
+
 type NetworkStatus struct {
 	// Connected is a flag that indicates whether this network is currently
 	// connected to the VM.
